from discord import Message, Client


class MessageHandler:

    async def __call__(self, client: Client, message: Message):
        if not self.accept(message):
            return
        await self.handle(client, message)

    def accept(self, message: Message) -> bool:
        raise NotImplementedError()

    async def handle(self, client: Client, message: Message):
        raise NotImplementedError()


from dataclasses import dataclass
from typing import Optional, Dict

from discord import Client as DiscordClient
from discord import Guild, CategoryChannel
from discord import Role, PermissionOverwrite

from discord_coworking.command.api import Result, UnundoableResult
from discord_coworking.command.guild import GuildResult, GuildCommand, GuildPredicate
from discord_coworking.command.overwrites import OverwritesMixin
from discord_coworking.command.predicate import ByAttributesPredicate
from discord_coworking.command.predicate import Predicate, ANY


class CategoryPredicate(ByAttributesPredicate[CategoryChannel]):

    @classmethod
    def of(cls, category: CategoryChannel):
        return cls.create(id=category.id)


@dataclass()
class CategoryCreated(GuildResult):
    category: Predicate[CategoryChannel] = ANY

    @classmethod
    def create(cls, guild: Guild, category: CategoryChannel):
        return cls(
            guild=GuildPredicate.of(guild),
            category=CategoryPredicate.of(category),
        )

    async def undo_on_guild(self, client: DiscordClient, guild: Guild):
        for category in self.category.filter(guild.categories):
            await category.delete()


@dataclass()
class CreateCategory(GuildCommand, OverwritesMixin):
    name: str = None
    overwrites: Optional[Dict[Predicate[Role], PermissionOverwrite]] = None

    async def do_on_guild(self, client: DiscordClient, guild: Guild) -> CategoryCreated:
        category = await guild.create_category(
            name=self.name,
            overwrites=self.get_overwrites(guild),
        )
        return CategoryCreated.create(guild, category)


class DeleteCategory(GuildCommand):
    category: Predicate[CategoryChannel] = ANY

    async def do_on_guild(self, client: DiscordClient, guild: Guild) -> Result:
        for category in self.category.filter(guild.categories):
            await category.delete()
        return UnundoableResult.create(self)



import logging
from dataclasses import dataclass, field
from logging import Logger
from typing import Sequence

from discord import Client as DiscordClient

from discord_coworking.command.api import Command, Result

__logger__: Logger = logging.getLogger(__name__)


@dataclass()
class BatchResult(Result):
    results: Sequence[Result] = field(default_factory=list)
    logger: Logger = field(default=__logger__)

    async def undo(self, client: DiscordClient):
        for result in self.results:
            try:
                await result.undo(client)
            except Exception as ex:
                self.logger.error(exc_info=ex)


@dataclass()
class BatchCommand(Command):
    commands: Sequence[Command] = field(default_factory=list)

    async def do(self, client: DiscordClient) -> Result:
        results = []
        for command in self.commands:
            results.append(await command.do(client))
        return BatchResult(results[::-1])

    @classmethod
    def create(cls, *commands: Command) -> 'BatchCommand':
        return cls(commands)



from dataclasses import dataclass
from dataclasses import field
from typing import Optional

from discord import Client as DiscordClient
from discord import Guild, Role, Color, Permissions

from discord_coworking.command.api import Result
from discord_coworking.command.api import UnundoableResult
from discord_coworking.command.guild import GuildResult, GuildCommand, GuildPredicate
from discord_coworking.command.predicate import ByAttributesPredicate
from discord_coworking.command.predicate import Predicate, ANY


class RolePredicate(ByAttributesPredicate[Role]):

    @classmethod
    def of(cls, role: Role):
        return cls.create(id=role.id)


class DefaultRolePredicate(Predicate[Role]):

    def __call__(self, e: Role) -> bool:
        return e.is_default()


class AdminRolePredicate(Predicate[Role]):

    def __call__(self, e: Role) -> bool:
        return e.permissions.administrator


DEFAULT_ROLE = DefaultRolePredicate()
NOT_DEFAULT_ROLE = -DEFAULT_ROLE
MANAGED_ROLE = RolePredicate.create(managed=True)
NOT_MANAGED_ROLE = -MANAGED_ROLE
ADMIN_ROLE = AdminRolePredicate()
NOT_ADMIN_ROLE = -ADMIN_ROLE


@dataclass()
class RoleCreated(GuildResult):
    role: Predicate[Role] = ANY

    @classmethod
    def create(cls, guild: Guild, role: Role):
        return cls(
            guild=GuildPredicate.of(guild),
            role=RolePredicate.of(role),
        )

    async def undo_on_guild(self, client: DiscordClient, guild: Guild):
        for role in filter(self.role, guild.roles):
            await role.delete()


@dataclass()
class CreateRole(GuildCommand):
    name: str = None
    permissions: Permissions = field(default_factory=Permissions.none)
    color: Color = field(default_factory=Color.default)
    hoist: bool = False
    mentionable: bool = False
    reason: Optional[str] = None

    async def do_on_guild(self, client: DiscordClient, guild: Guild) -> RoleCreated:
        role = await guild.create_role(
            name=self.name,
            permissions=self.permissions,
            colour=self.color,
            hoist=self.hoist,
            mentionable=self.mentionable,
            reason=self.reason,
        )
        return RoleCreated.create(guild, role)


@dataclass()
class DeleteRole(GuildCommand):
    role: RolePredicate = NOT_DEFAULT_ROLE & NOT_MANAGED_ROLE & NOT_ADMIN_ROLE

    async def do_on_guild(self, client: DiscordClient, guild: Guild) -> Result:
        for role in self.role.filter(guild.roles):
            await role.delete()
        return UnundoableResult.create(self)



import asyncio
import tempfile
import emoji
import shlex
import json
import multiprocessing
from concurrent.futures import ThreadPoolExecutor
import argparse
import asyncio
import io
import logging.config
import os
import pickle
from asyncio.tasks import Task

import discord

from discord.message import Message

from discord_coworking.bot.handler import MessageHandlerNode
from discord_coworking.command.api import Result
from discord_coworking.command.category import DeleteCategory
from discord_coworking.command.coworking import CreateOrganization
from discord_coworking.command.guild import GuildPredicate
from discord_coworking.command.predicate import ANY
from discord_coworking.command.role import DeleteRole
from discord_coworking.command.text_channel import DeleteTextChannel
from discord_coworking.command.voice_channel import DeleteVoiceChannel
from discord_coworking.bot import CoworkingBot, StartByMentioningMe
from discord_coworking.bot.handler.run_in_docker import RunCodeInDocker

def create_parser(bot_command=True, token_param=True, log_level_param=True,
                  client_factory_param=True) -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog='discord-coworking',
        description='Discord coworking server management toolkit',
    )
    if token_param:
        parser.add_argument(
            '--token',
            help='Discord Bot Token, environment DISCORD_TOKEN',
            **default_env('DISCORD_TOKEN', str, None),
        )
    parser.add_argument(
        '--guild',
        help='Guild ID',
        **default_env('GUILD_ID', lambda it: GuildPredicate.create(id=int(it)), ANY)
    )
    if log_level_param:
        parser.add_argument(
            '--log-level',
            help='Log level',
            choices=('DEBUG', 'INFO', 'WARN', 'WARNING', 'ERROR', 'CRITICAL',),
            default='INFO',
        )
    if client_factory_param:
        parser.set_defaults(
            client_factory=lambda : discord.Client(shard_id=0, shard_count=1),
        )
    subparsers = parser.add_subparsers(
        title='sub command',
    )
    undo = subparsers.add_parser(
        'undo',
    )
    undo.add_argument(
        '-I', '-i', '--input',
        help='Serialized result file',
        type=argparse.FileType('rb'),
    )
    undo.set_defaults(command=undo_command)
    if bot_command:
        bot = subparsers.add_parser(
            'bot',
        )
        bot.set_defaults(
            client_factory=CoworkingBot,
            command=run_bot,
        )
    create_organization_commands = {
        'create-open-organization': CreateOrganization.open_organization,
        'create-private-organization': CreateOrganization.private_organization,
    }
    for name, factory in create_organization_commands.items():
        create_organization = subparsers.add_parser(
            name,
        )
        create_organization.add_argument(
            '-o', '-O', '--output',
            dest='output',
            default='a.out',
            type=argparse.FileType('wb+'),
        )
        create_organization.add_argument(
            '--name',
            required=True,
        )
        create_organization.set_defaults(
            command=discord_command,
            command_factory=factory,
        )
    delete_commands = {
        'delete-category': DeleteCategory,
        'delete-voice-channel': DeleteVoiceChannel,
        'delete-text-channel': DeleteTextChannel,
        'delete-role': DeleteRole,
    }
    for name, factory in delete_commands.items():
        create_organization = subparsers.add_parser(
            name,
        )
        create_organization.add_argument(
            '-O', '-o', '--output',
            dest='output',
            default='a.out',
            type=argparse.FileType('wb+'),
        )
        create_organization.set_defaults(
            command=discord_command,
            command_factory=factory,
        )
    list_commands = {
        'list-guild': None,
    }
    for name, _ in list_commands.items():
        command = subparsers.add_parser(
            name,
        )
        command.set_defaults(
            command=list_command,
        )
    return parser


def config_log(log_level):
    logging.basicConfig(level=log_level)
    logging.info(f'Log level set to {log_level}')


async def run(token: str, client_factory, command, log_level, **params):
    config_log(log_level)
    client = client_factory()
    loop = asyncio.get_event_loop()
    client_task = loop.create_task(client.start(token))
    await client.wait_until_ready()
    await command(client, **params, client_task=client_task)


async def list_command(client: discord.Client, **params):
    for guild in client.guilds:
        print(f'"{guild.name}" ({guild.id})')


async def print_handler(client, message: Message):
    print(message)
    print(message.mentions)
    print(message.content)
    print(message.clean_content)


async def reply_handler(client, message: Message):
    await message.reply(message.clean_content)


def error(err):
    raise Exception(err)


def exit(status=0, message=None):
    if status == 0:
        return
    raise Exception(message)


class RunAsCommand():
    tool_emoji = emoji.EMOJI_ALIAS_UNICODE[':hammer_and_wrench:']
    loading_emoji = emoji.EMOJI_ALIAS_UNICODE[':hourglass:']
    success_emoji = emoji.EMOJI_ALIAS_UNICODE[':white_check_mark:']
    error_emoji = emoji.EMOJI_ALIAS_UNICODE[':x:']

    def __init__(self):
        self.parser = create_parser(bot_command=False, token_param=False, log_level_param=False,
                                    client_factory_param=False)
        self.parser.error = error
        self.parser.exit = exit

    async def __call__(self, client: discord.Client, message: Message):
        if not message.author.guild_permissions.administrator:
            await message.reply("Sorry I can't help you with that, ask to an administrator")
            return
        await message.add_reaction(self.tool_emoji)
        args = shlex.split(message.clean_content)
        try:
            input_file = None
            for attachment in message.attachments:
                _, input_file = tempfile.mkstemp()
                with open(input_file, 'wb+') as attachment_output:
                    attachment_output.write(await attachment.read())
                break
            if input_file:
                args += [f'--input={input_file}']
            namespace = self.parser.parse_args([f"--guild={message.guild.id}"] + args[1:])
            params = vars(namespace)
            command = params.pop('command')
            await message.add_reaction(self.loading_emoji)
            await command(client=client, **params)
            files = []
            if 'output' in namespace:
                namespace.output.close()
                file = os.path.abspath(namespace.output.name)
                files.append(discord.File(fp=file, filename=os.path.basename(file)))
            await message.reply('=)', files=files)
            await message.add_reaction(self.success_emoji)
        except KeyError:
            await message.reply(self.parser.format_help())
        except Exception as ex:
            await message.add_reaction(self.error_emoji)
            await message.reply(str(ex))
        finally:
            await message.remove_reaction(self.loading_emoji, client.user)


async def run_bot(client: CoworkingBot, client_task: Task, **params):
    client.message_handler.handlers.extend([
        RunCodeInDocker(),
        MessageHandlerNode(RunAsCommand(), StartByMentioningMe(client)),
    ])
    await client_task


async def undo_command(client: discord.Client, input: io.BufferedReader, **params):
    undoable = pickle.load(input)
    if not isinstance(undoable, (Result,)):
        raise Exception('The input file is not a result')
    await undoable.undo(client)


async def discord_command(client: discord.Client, command_factory, output: io.BufferedWriter, **parameters):
    command = command_factory(**parameters)
    logging.info(f'Running command {command}')
    pickle.dump(await command.do(client), output)


def default_env(name, type, default):
    env_value = os.environ.get(name, None)
    return {
        'default': default if env_value is None else type(env_value),
        'required': env_value is None and default is None,
        'type': type,
    }


async def main():
    parser = create_parser()
    namespace = parser.parse_args()
    return await run(**vars(namespace))


try:
    cpu_count = multiprocessing.cpu_count()
    thread_pool = ThreadPoolExecutor(cpu_count)
    loop = asyncio.new_event_loop()
    loop.set_default_executor(thread_pool)
    loop.run_until_complete(main())
except KeyboardInterrupt:
    asyncio.get_event_loop().close()



from typing import Optional, Dict

from discord import Role, PermissionOverwrite, Guild

from discord_coworking.command.predicate import Predicate


class OverwritesMixin:
    overwrites: Optional[Dict[Predicate[Role], PermissionOverwrite]] = None

    def get_overwrites(self, guild: Guild) -> Optional[Dict[Role, PermissionOverwrite]]:
        if self.overwrites is None:
            return None
        return dict(
            (role_id.any(guild.roles), perms)
            for role_id, perms in self.overwrites.items()
        )



from dataclasses import dataclass
from typing import Optional, Dict

from discord import Client as DiscordClient
from discord import Guild, TextChannel, CategoryChannel, Role, PermissionOverwrite

from discord_coworking.command.api import Result, UnundoableResult
from discord_coworking.command.guild import GuildResult, GuildCommand, GuildPredicate
from discord_coworking.command.overwrites import OverwritesMixin
from discord_coworking.command.predicate import ANY, NONE
from discord_coworking.command.predicate import ByAttributesPredicate, Predicate


class TextChannelPredicate(ByAttributesPredicate[TextChannel]):

    @classmethod
    def of(cls, text_channel: TextChannel):
        return cls.create(id=text_channel.id)


@dataclass()
class TextChannelCreated(GuildResult):
    text_channel: Predicate[TextChannel] = ANY

    @classmethod
    def create(cls, guild: Guild, text_channel: TextChannel):
        return cls(
            guild=GuildPredicate.of(guild),
            text_channel=TextChannelPredicate.of(text_channel),
        )

    async def undo_on_guild(self, client: DiscordClient, guild: Guild):
        for text_channel in self.text_channel.filter(guild.text_channels):
            await text_channel.delete()


@dataclass()
class CreateTextChannel(GuildCommand, OverwritesMixin):
    name: str = None
    topic: Optional[str] = None
    category: Predicate[CategoryChannel] = NONE
    overwrites: Optional[Dict[Predicate[Role], PermissionOverwrite]] = None

    async def do_on_guild(self, client: DiscordClient, guild: Guild) -> TextChannelCreated:
        text_channel = await guild.create_text_channel(
            name=self.name,
            topic=self.topic,
            category=self.category.any(guild.categories),
            overwrites=self.get_overwrites(guild),
        )
        return TextChannelCreated.create(guild, text_channel)


class DeleteTextChannel(GuildCommand):
    text_channel: Predicate[TextChannel] = ANY

    async def do_on_guild(self, client: DiscordClient, guild: Guild) -> Result:
        for text_channel in self.text_channel.filter(guild.text_channels):
            await text_channel.delete()
        return UnundoableResult.create(self)



import logging
from dataclasses import dataclass, field
from logging import Logger
from typing import Optional, TypeVar, Generic

from discord import Client as DiscordClient

__logger__: Logger = logging.getLogger(__name__)


class Result:

    async def undo(self, client: DiscordClient):
        raise NotImplementedError()


class Command:

    async def do(self, client: DiscordClient) -> Optional[Result]:
        raise NotImplementedError()


@dataclass
class UnundoableResult(Result):
    command: Command = None
    logger: Logger = field(default=__logger__)

    async def undo(self, client: DiscordClient):
        self.logger.warning(f'The command {self.command} cannot be undone')

    @classmethod
    def create(cls, command: Command):
        return cls(command)


C = TypeVar('C')


@dataclass()
class UndoableResult(Generic[C], Result):
    command: C = None
    logger: Logger = field(default=__logger__)

    async def undo(self, client: DiscordClient):
        raise NotImplementedError()



import asyncio
from dataclasses import dataclass
from typing import Callable, Awaitable, List

from discord import Client
from discord.message import Message

from discord_coworking.bot.api import MessageHandler
from discord_coworking.command.predicate import Predicate, ANY


@dataclass()
class MessageHandlerNode(MessageHandler):
    handler: Callable[[Client, Message], Awaitable]
    predicate: Predicate[Message] = ANY

    def accept(self, message: Message) -> bool:
        return self.predicate(message)

    async def handle(self, client: Client, message: Message):
        await self.handler(client, message)


@dataclass()
class MessageHandlerChain(MessageHandler):
    handlers: List[MessageHandler]
    predicate: Predicate[Message] = ANY

    def accept(self, message: Message) -> bool:
        return self.predicate(message)

    async def handle(self, client: Client, message: Message):
        for handler in self.handlers:
            if handler.accept(message):
                asyncio.create_task(handler.handle(client, message))
